/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is Mozilla Communicator client code, 
 * released March 31, 1998. 
 *
 * The Initial Developer of the Original Code is Netscape Communications 
 * Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s): 
 *     Daniel Veditz <dveditz@netscape.com>
 *     Samir Gehani <sgehani@netscape.com>
 *     Mitch Stoltz <mstoltz@netscape.com>
 *     Jeroen Dobbelaere <jeroen.dobbelaere@acunia.com>
 *     Alfred Kayser <alfredkayser@nl.ibm.com>
 */

/* 
 * This module implements a simple archive extractor for the PKZIP format.
 *
 * The underlying nsZipArchive is NOT thread-safe. Do not pass references
 * or pointers to it across thread boundaries.
 */


#ifndef STANDALONE

#include "nsWildCard.h"
#include "nscore.h"
#include "prmem.h"
#include "prio.h"
#include "plstr.h"
#include "prlog.h"
#define ZFILE_CREATE    PR_WRONLY | PR_CREATE_FILE
#define READTYPE  PRInt32
#include "zlib.h"
#include "nsISupportsUtils.h"
#include "nsRecyclingAllocator.h"

/**
 * Globals
 *
 * Global allocator used with zlib. Destroyed in module shutdown.
 */
#define NBUCKETS 6
#define BY4ALLOC_ITEMS 320
nsRecyclingAllocator *gZlibAllocator = NULL;

#else /* STANDALONE */

#ifdef XP_WIN
#include "windows.h"
#endif

#undef MOZILLA_CLIENT       // undoes prtypes damage in zlib.h
#define ZFILE_CREATE  "wb"
#define READTYPE  PRUint32
#include "zlib.h"
#undef PR_PUBLIC_API
#include "zipstub.h"

#ifdef XP_MAC
#include <string.h>
#include <stdlib.h>

char * strdup(const char *src);
char * strdup(const char *src)
{
    long len = strlen(src);
    char *dup = (char *)malloc(len+1 * sizeof(char));
    memcpy(dup, src, len+1);
    return dup;
}
#endif

#endif /* STANDALONE */

#ifdef XP_UNIX
    #include <sys/stat.h>
    #include <limits.h>
    #include <unistd.h>
#elif defined(XP_WIN) || defined(XP_OS2)
    #include <io.h>
#endif

#include "zipfile.h"
#include "zipstruct.h"
#include "nsZipArchive.h"

static PRUint16 xtoint(unsigned char *ii);
static PRUint32 xtolong(unsigned char *ll);

/*---------------------------------------------
 * C API wrapper for nsZipArchive
 *--------------------------------------------*/

#ifdef STANDALONE

/**
 * ZIP_OpenArchive
 *
 * opens the named zip/jar archive and returns a handle that
 * represents the archive in other ZIP_ calls.
 *
 * @param   zipname   archive filename
 * @param   hZip      receives handle if archive opened OK
 * @return  status code
 */
PR_PUBLIC_API(PRInt32) ZIP_OpenArchive( const char * zipname, void** hZip )
{
  PRInt32 status;

  /*--- error check args ---*/
  if ( hZip == 0 )
    return ZIP_ERR_PARAM;

  /*--- NULL output to prevent use by bozos who don't check errors ---*/
  *hZip = 0;

  /*--- create and open the archive ---*/
  nsZipArchive* zip = new nsZipArchive();
  if ( zip == 0 )
    return ZIP_ERR_MEMORY;

  status = zip->OpenArchive(zipname);

  if ( status == ZIP_OK )
    *hZip = NS_STATIC_CAST(void*,zip);
  else
      delete zip;

  return status;
}



/**
 * ZIP_TestArchive
 *
 * Tests the integrity of this open zip archive by extracting each 
 * item to memory and performing a CRC check.
 *
 * @param   hZip      handle obtained from ZIP_OpenArchive
 * @return  status code (success indicated by ZIP_OK)
 */
PR_PUBLIC_API(PRInt32) ZIP_TestArchive( void *hZip )
{
  /*--- error check args ---*/
  if ( hZip == 0 )
    return ZIP_ERR_PARAM;
   
  nsZipArchive* zip = NS_STATIC_CAST(nsZipArchive*,hZip);
  if ( zip->kMagic != ZIP_MAGIC )
    return ZIP_ERR_PARAM;   /* whatever it is isn't one of ours! */

  /*--- test the archive ---*/
  return zip->Test(NULL);
}


/**
 * ZIP_CloseArchive
 *
 * closes zip archive and frees memory
 * @param   hZip  handle obtained from ZIP_OpenArchive
 * @return  status code
 */
PR_PUBLIC_API(PRInt32) ZIP_CloseArchive( void** hZip )
{
  /*--- error check args ---*/
  if ( hZip == 0 || *hZip == 0 )
    return ZIP_ERR_PARAM;

  nsZipArchive* zip = NS_STATIC_CAST(nsZipArchive*,*hZip);
  if ( zip->kMagic != ZIP_MAGIC )
    return ZIP_ERR_PARAM;   /* whatever it is isn't one of ours! */

  /*--- close the archive ---*/
  *hZip = 0;
  delete zip;

  return ZIP_OK;
}



/**
 * ZIP_ExtractFile
 *
 * extracts named file from an opened archive
 *
 * @param   hZip      handle obtained from ZIP_OpenArchive
 * @param   filename  name of file in archive
 * @param   outname   filename to extract to
 */
PR_PUBLIC_API(PRInt32) ZIP_ExtractFile( void* hZip, const char * filename, const char * outname )
{
  /*--- error check args ---*/
  if ( hZip == 0 )
    return ZIP_ERR_PARAM;

  nsZipArchive* zip = NS_STATIC_CAST(nsZipArchive*,hZip);
  if ( zip->kMagic != ZIP_MAGIC )
    return ZIP_ERR_PARAM;   /* whatever it is isn't one of ours! */

  /*--- extract the file ---*/
  return zip->ExtractFile( filename, outname );
}



/**
 * ZIP_FindInit
 *
 * Initializes an enumeration of files in the archive
 *
 * @param   hZip      handle obtained from ZIP_OpenArchive
 * @param   pattern   regexp to match files in archive, the usual shell expressions.
 *                    NULL pattern also matches all files, faster than "*"
 */
PR_PUBLIC_API(void*) ZIP_FindInit( void* hZip, const char * pattern )
{
  /*--- error check args ---*/
  if ( hZip == 0 )
    return 0;

  nsZipArchive* zip = NS_STATIC_CAST(nsZipArchive*,hZip);
  if ( zip->kMagic != ZIP_MAGIC )
    return 0;   /* whatever it is isn't one of ours! */

  /*--- initialize the pattern search ---*/
  return zip->FindInit( pattern );
}



/**
 * ZIP_FindNext
 *
 * Puts the next name in the passed buffer. Returns ZIP_ERR_SMALLBUF when
 * the name is too large for the buffer, and ZIP_ERR_FNF when there are no 
 * more files that match the pattern
 *
 * @param   hFind     handle obtained from ZIP_FindInit
 * @param   outbuf    buffer to receive next filename
 * @param   bufsize   size of allocated buffer
 */
PR_PUBLIC_API(PRInt32) ZIP_FindNext( void* hFind, char * outbuf, PRUint16 bufsize )
{
  PRInt32 status;

  /*--- error check args ---*/
  if ( hFind == 0 )
    return ZIP_ERR_PARAM;

  nsZipFind* find = NS_STATIC_CAST(nsZipFind*,hFind);
  if ( find->kMagic != ZIPFIND_MAGIC )
    return ZIP_ERR_PARAM;   /* whatever it is isn't one of ours! */

  /*--- return next filename file ---*/
  nsZipItem* item;
  status = find->GetArchive()->FindNext( find, &item );
  if ( status == ZIP_OK )
    status = item->GetName(outbuf, bufsize);

  return status;
}



/**
 * ZIP_FindFree
 *
 * Releases allocated memory associated with the find token
 *
 * @param   hFind     handle obtained from ZIP_FindInit
 */
PR_PUBLIC_API(PRInt32) ZIP_FindFree( void* hFind )
{
  /*--- error check args ---*/
  if ( hFind == 0 )
    return ZIP_ERR_PARAM;

  nsZipFind* find = NS_STATIC_CAST(nsZipFind*,hFind);
  if ( find->kMagic != ZIPFIND_MAGIC )
    return ZIP_ERR_PARAM;   /* whatever it is isn't one of ours! */

  /* free the find structure */
  return find->GetArchive()->FindFree( find );
}

#if defined XP_WIN
void ProcessWindowsMessages()
{
  MSG msg;

  while(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
}
#endif /* XP_WIN */

#else /* STANDALONE */

//***********************************************************
// Allocators for use with zlib
//
// These are allocators that are performance tuned for
// use with zlib. Our use of zlib for every file we read from
// the jar file when running navigator, we do these allocation.
// alloc 24
// alloc 64
// alloc 11520
// alloc 32768
// alloc 1216 [304x4] max
// alloc 76   [19x4]
// free  76   [19x4]
// alloc 1152 [288x4]
// free  1152 [288x4]
// free  1216 [304x4]
// alloc 28
// free  28
// free  32768
// free  11520
// free  64
// free  24
//
// The pool will allocate these as:
//
//          32,768
//          11,520
//           1,280 [320x4] - shared by first x4 alloc, 28
//           1,280 [320x4] - shared by second and third x4 alloc
//              64
//              24
//          ------
//          46,936
//
// And almost all of the file reads happen serially. Hence this
// allocator tries to keep one set of memory needed for one file around
// and reused the same blocks for other file reads.
//
// The interesting question is when should be free this ?
// - memory pressure should be one.
// - after startup of navigator
// - after startup of mail
// In general, this allocator should be enabled before
// we startup and disabled after we startup if memory is a concern.
//***********************************************************

PR_STATIC_CALLBACK(void *)
zlibAlloc(void *opaque, uInt items, uInt size)
{
  nsRecyclingAllocator *zallocator = (nsRecyclingAllocator *)opaque;
  if (zallocator) {
    // Bump up x4 allocations
    PRUint32 realitems = items;
    if (size == 4 && items < BY4ALLOC_ITEMS)
      realitems = BY4ALLOC_ITEMS;
     return zallocator->Calloc(realitems, size);
  }
  else
    return calloc(items, size);
}

PR_STATIC_CALLBACK(void)
zlibFree(void *opaque, void *ptr)
{
  nsRecyclingAllocator *zallocator = (nsRecyclingAllocator *)opaque;
  if (zallocator)
    zallocator->Free(ptr);
  else
    free(ptr);
  return;
}
#endif /* STANDALONE */

//***********************************************************
//      nsZipArchive  --  public methods
//***********************************************************


//---------------------------------------------
//  nsZipArchive::OpenArchive
//---------------------------------------------
PRInt32 nsZipArchive::OpenArchive( const char * aArchiveName )
{
  //-- validate arguments
  if ( aArchiveName == 0 || *aArchiveName == '\0') 
    return ZIP_ERR_PARAM;

  //-- not allowed to do two opens on the same object!
  if ( mMap != 0 )
    return ZIP_ERR_GENERAL;

  //-- open the physical file
  PRFileDesc *mFd = PR_Open( aArchiveName, PR_RDONLY, 0 );
  if ( mFd == 0 )
    return ZIP_ERR_DISK;

  //-- get table of contents for archive
  PRInt32 status = BuildFileList(mFd);

  PR_Close(mFd);

  return status;
}

PRInt32 nsZipArchive::OpenArchiveWithFileDesc(PRFileDesc* fd)
{
  //-- validate arguments
  if (fd == 0) 
    return ZIP_ERR_PARAM;

  //-- not allowed to do two opens on the same object!
  if ( mMap != 0 )
    return ZIP_ERR_GENERAL;

  //-- get table of contents for archive
  return BuildFileList(fd);
}

//---------------------------------------------
//  nsZipArchive::Test
//---------------------------------------------
PRInt32 nsZipArchive::Test(const char *aEntryName)
{
  PRInt32 rv = ZIP_OK;
  nsZipItem *currItem = 0;

  if (aEntryName) // only test specified item
  {
    currItem = GetFileItem(aEntryName);
    if(!currItem)
      return ZIP_ERR_FNF;

    rv = TestItem(currItem);
  }
  else // test all items in archive
  {
    nsZipFind *iterator = FindInit( NULL );
    if (!iterator)
      return ZIP_ERR_GENERAL;

    // iterate over items in list
    while (ZIP_OK == FindNext(iterator, &currItem)) 
    {
      rv = TestItem(currItem);

      // check if crc check failed
      if (rv != ZIP_OK)
        break;

#if defined STANDALONE && defined XP_WIN
      ProcessWindowsMessages();
#endif
    }
  
    FindFree(iterator);
  }

  return rv;
}

//---------------------------------------------
//  nsZipArchive::CloseArchive
//---------------------------------------------
PRInt32 nsZipArchive::CloseArchive()
{
  PRStatus rv;

  // close the file if open
  if ( mMap != 0 ) {
    rv = PR_MemUnmap(mMap, ziplen);
    PR_ASSERT(rv == PR_SUCCESS);
    mMap = 0;
  }
  if (fMap != null) {
    rv = PR_CloseFileMap(fMap);
    PR_ASSERT(rv == PR_SUCCESS);
    fMap = null;
  }
  PR_FREEIF(mItems);

  return ZIP_OK;
}

//---------------------------------------------
// nsZipArchive::GetItem
//---------------------------------------------
PRInt32 nsZipArchive::GetItem( const char * aFilename, nsZipItem **result)
{
  //-- Parameter validity check
  if (aFilename == 0)
    return ZIP_ERR_PARAM;

  //-- find file information
  nsZipItem *item = GetFileItem( aFilename );
  if ( item == 0 )
    return ZIP_ERR_FNF;

  *result = item; // Return a pointer to the struct
  return ZIP_OK;
}


//---------------------------------------------
// nsZipArchive::ReadInit
//---------------------------------------------
PRInt32 nsZipArchive::ReadInit(const char* aName, nsZipRead* aRead)
{
  PRInt32 result = ZIP_OK;
  char* buf;

  //-- Parameter validity check
  if (zipEntry == 0 || aRead == 0)
    return ZIP_ERR_PARAM;

  //-- find item
  nsZipItem* item = GetFileItem(aName);
  if (!item)
    return ZIP_ERR_FNF; 

  //-- Initialize nsZipRead object
  aRead->Init(this, item);
  switch(item->GetCompression())
  {
    case DEFLATED:
      //-- Read the item into memory
      //   Inflate if necessary and save in mInflatedFileBuffer
      //   for sequential reading.
      //   (nsJAR needs the whole file in memory before passing it on)
      buf = (char*)PR_Malloc(item->GetRealSize());
      if (!buf) return ZIP_ERR_MEMORY;
      result = InflateToBuffer(item, buf);
      break;

    case STORED:
      // Just set the pointer to the piece in mmap.
      buf = mMap + GetOffset( item );   
      break;

    default:
      return ZIP_ERR_UNSUPPORTED;
  }

  if (result == ZIP_OK)
    //-- verify crc32
    PRUint32 calculatedCRC = crc32(0L, Z_NULL, 0);
    calculatedCRC = crc32( calculatedCRC,(const unsigned char*)buf, item->GetRealSize());
    if (calculatedCRC != item->GetCRC32())
      result = ZIP_ERR_CORRUPT;
    else
      aRead->mFileBuffer = buf; // All clear to be used
  }
  return result;
}
  
//------------------------------------------
// nsZipArchive::Read
//------------------------------------------
PRInt32 nsZipArchive::Read(nsZipRead* aRead, char* aOutBuf,
                           PRUint32 aCount, PRUint32* aBytesRead)
{
  //-- Copies data from aRead->mFileBuffer
  //-- Parameter check
  if (aBytesRead == 0 || aRead == 0 || aOutBuf == 0 ||
      aRead->mArchive != this)
    return ZIP_ERR_PARAM;

  if(!aRead->mItem || !aRead->mFileBuffer)
     return ZIP_ERR_GENERAL;

  //-- Set up the copy
  PRUint32 bigBufSize = aRead->mItem->GetRealSize();
  PRUint32 curPos = aRead->mCurPos;
  *aBytesRead  = (curPos + aCount) < bigBufSize ? aCount : bigBufSize - curPos;
  char* src = aRead->mFileBuffer + curPos;

  //-- Do the copy and record number of bytes copied
  memcpy(aOutBuf, src, *aBytesRead);
  aRead->mCurPos += *aBytesRead;

  return ZIP_OK;
}

//---------------------------------------------
// nsZipArchive::Available
//---------------------------------------------
PRUint32 nsZipArchive::Available(nsZipRead* aRead)
{
  if (aRead == 0)
    return 0;

  nsZipItem* item = aRead->mItem;

  if (item->GetCompression() == DEFLATED)
    return item->GetRealSize() - aRead->mCurPos;
  else
    return item->GetSize() - aRead->mCurPos;
}

//---------------------------------------------
// nsZipArchive::ExtractFile
//---------------------------------------------
PRInt32 
nsZipArchive::ExtractFile(const char * inName, const char* outName)
{
  PRFileDesc* fOut = PR_Open(outName, ZFILE_CREATE, 0644);
  if (fOut == 0)
    return ZIP_ERR_DISK;

  nsZipItem *item = GetFileItem( inName );
  PRInt32 status = ExtractToFileDesc(item, fOut);
  PR_Close(fOut);

  if (status != ZIP_OK) 
  {
    PR_Delete(aOutname);
  }
#if defined(XP_UNIX)
  else 
  {
    if (item->IsSymLink())
    {
       status = ResolveSymlink(outName);
    }
    //-- set extracted file permissions
    chmod(outName, item->GetMode());
  }
#endif
  return status;
}

   
//---------------------------------------------
// nsZipArchive::FindInit
//---------------------------------------------
nsZipFind* nsZipArchive::FindInit( const char * aPattern )
{
  PRBool  regExp = PR_FALSE;
  char*   pattern = 0;

  // validate the pattern
  if ( aPattern )
  {
    switch (NS_WildCardValid( (char*)aPattern ))
    {
      case INVALID_SXP:
        return 0;

      case NON_SXP:
        regExp = PR_FALSE;
        break;

      case VALID_SXP:
        regExp = PR_TRUE;
        break;

      default:
        // undocumented return value from RegExpValid!
        PR_ASSERT( PR_FALSE );
        return 0;
    }

    pattern = PL_strdup( aPattern );
    if ( !pattern )
      return 0;
  }

  return new nsZipFind( this, pattern, regExp );
}



//---------------------------------------------
// nsZipArchive::FindNext
//---------------------------------------------
PRInt32 nsZipArchive::FindNext( nsZipFind* aFind, nsZipItem** aResult)
{
  PRInt32    status;
  PRBool     found  = PR_FALSE;
  PRUint16   slot   = aFind->mSlot;
  nsZipItem* item   = aFind->mItem;

  if ( aFind->mArchive != this )
    return ZIP_ERR_PARAM;

  // we start from last match, look for next
  while ( slot < ZIP_TABSIZE && !found )
  {
    if ( item != 0 ) 
      item = item->next;    // move to next in current chain
    else
      item = mFiles[slot];  // starting a new slot

    if ( item == 0 )
    { // no more in this chain, move to next slot
      ++slot;
      continue;
    }
    else if ( aFind->mPattern == 0 )  
      found = PR_TRUE;            // always match
    else {
      PRUint16 namelen = xtoint( item->central->filename_len );
      const char *name = item->central + ZIPCENTRAL_SIZE;
      if ( aFind->mRegExp ) {
        char *dup = (char *)malloc(namelen+1 * sizeof(char));
        if (dup) {  // If out of memory, just fail the match
          memcpy(dup, name, namelen+1);
          dup[namelen+1] = '\0';
          found = (NS_WildCardMatch( dup, aFind->mPattern, PR_FALSE ) == MATCH);
          free(dup);
        }
      } else {
        int patlen = strlen(aFind->mPattern);
#if defined(STANDALONE) && defined(XP_MAC)
        // simulate <regexp>* matches
        if (namelen >= patlen) {
#else
        if (namelen == patlen) {
#endif
          found = ( memcmp(name, aFind->mPattern, patlen) == 0 );
        }
      }
    }
  }

  if ( found ) 
  {
      *aResult = item;
      aFind->mSlot = slot;
      aFind->mItem = item;
      status = ZIP_OK;
  }
  else
    status = ZIP_ERR_FNF;

  return status;
}


//---------------------------------------------
// nsZipArchive::FindFree
//---------------------------------------------
PRInt32 nsZipArchive::FindFree( nsZipFind* aFind )
{
  if ( aFind->mArchive != this )
    return ZIP_ERR_PARAM;

  delete aFind;
  return ZIP_OK;
}


//***********************************************************
//      nsZipArchive  --  private implementation
//***********************************************************

//---------------------------------------------
//  nsZipArchive::BuildFileList
//---------------------------------------------
PRInt32 nsZipArchive::BuildFileList(PRFileDesc *fd)
{
  PRInt32   status = ZIP_OK;

  //-- get archive size using end pos
  ziplen = PR_Seek(fd, 0, PR_SEEK_END);
#ifdef STANDALONE
  if (ziplen  || ((ziplen = ftell(fd)) <= 0))
#else
  if (ziplen  <= 0)
#endif
    return ZIP_ERR_CORRUPT;

  PRInt64 sz64;
  LL_UI2L(sz64, ziplen);
  fMap = PR_CreateFileMap(fd, sz64, PR_PROT_READONLY);
  PR_ASSERT(fMap);
  mMap = PR_MemMap(fMap, LL_ZERO, ziplen);
  if (mMap == (void *) -1) {
    PR_CloseFileMap(fMap);
    fMap = null;
    return ZIP_ERR_DISK;
  }

  //-----------------------------------------------------------------------
  // locate the central directory via the End record
  //-----------------------------------------------------------------------
  PRUint32 pos;
  for (pos = ziplen-4; pos>0; pos--) {
    if (xtolong(mMap + pos) == ENDSIG) {
      ZipEnd *End = mMap + pos;
      //-- set pos to start of central directory
      pos = xtolong(End->offset_central_dir);
      break;
    }
  }
  if (pos == 0)
    status = ZIP_ERR_CORRUPT;

  //-------------------------------------------------------
  // count the central directory headers
  //-------------------------------------------------------
  PRUint32 p = pos;
  count = 0;
  while ( status == ZIP_OK )
  {
    PRUint32 sig = xtolong( mMap + p );
    if ( sig != CENTRALSIG ) {
      //-- we must be done or else archive is corrupt
      if ( sig != ENDSIG )
        status = ZIP_ERR_CORRUPT;
      break;
    }
    //-- make sure we've read enough
    if ( ziplen < p + ZIPCENTRAL_SIZE )
    {
      status = ZIP_ERR_CORRUPT;
      break;
    }
    ZipCentral* central = (ZipCentral*)(mMap + p);
    PRUint32 namelen = xtoint( central->filename_len );
    PRUint32 extralen = xtoint( central->extrafield_len );
    PRUint32 commentlen = xtoint( central->commentfield_len );

    //-- set position to start of next ZipCentral record
    p += ZIPCENTRAL_SIZE + namelen + extralen + commentlen;
    count ++;   
  }

  //-------------------------------------------------------
  // Index the central directory headers
  //-------------------------------------------------------
  mItems = PR_Malloc(sizeof(nsZipItem) * count);
  p = pos;
  for (PRInt32 i=0; i<count; i++) {
    ZipCentral* central = (ZipCentral*)(mMap + p);
    PRUint32 namelen = xtoint( central->filename_len );
    PRUint32 extralen = xtoint( central->extrafield_len );
    PRUint32 commentlen = xtoint( central->commentfield_len );
    p += ZIPCENTRAL_SIZE;

    //-- add item to file table
    nsZipItem* item = mItems+i;
    PRUint32 hash = HashName( p, namelen );
    item->next = mFiles[hash];
    mFiles[hash] = item;
    
    p += namelen + extralen + commentlen;
  }

#if defined(DEBUG)
  if(status != ZIP_OK) {
    const char* msgs[] = { "ZIP_OK", "ZIP_ERR_GENERAL", "ZIP_ERR_MEMORY", "ZIP_ERR_DISK", 
        "ZIP_ERR_CORRUPT", "ZIP_ERR_PARAM", "ZIP_ERR_FNF", "ZIP_ERR_UNSUPPORTED", 
        "ZIP_ERR_SMALLBUF", "UNKNOWN" };
    printf("nsZipArchive::BuildFileList  status = %d '%s'\n", 
        (int)status, msgs[(status <= 0 && status >= -8) ? -status : 9]);
  }
#endif

  return status;
}
  
//---------------------------------------------
// nsZipArchive::GetFileItem
//---------------------------------------------
nsZipItem*  nsZipArchive::GetFileItem( const char * aName )
{
  PR_ASSERT( aName != 0 );

  int aLen = strlen(aName);
  nsZipItem* item = mFiles[ HashName(aName, aLen) ];

  for ( ; item != 0; item = item->next )
  {
    PRUint16 namelen = xtoint( item->central->filename_len );
    if ( namelen == aLen ) 
    {
      if ( 0 == memcmp( aName, item->central + ZIPCENTRAL_SIZE, namelen) ) 
        break; //-- found it
    }
  }

  return item;
}


//---------------------------------------------
// nsZipArchive::HashName
//---------------------------------------------
PRUint32 nsZipArchive::HashName( const char* aName, int len )
{
  PRUint32 val = 0;

  PR_ASSERT( aName != 0 );

  for (int i=0; i<len; i++) {  
    val = val*37 + (PRUint8)(aName[i]);
  }

  return (val % ZIP_TABSIZE);
}

//---------------------------------------------
// nsZipArchive::GetOffset
//---------------------------------------------
PRInt32  nsZipArchive::GetOffset(const nsZipItem* aItem)
{
  PR_ASSERT (aItem);

  PRUint32 offset = xtolong(aItem->central->localhdr_offset);
  ZipLocal *Local = mMap + offset;
  offset += ZIPLOCAL_SIZE + xtoint( Local->filename_len ) + xtoint( Local->extrafield_len );

  return offset;
}

/*
 * nsZipArchive::ExtractToFileDesc
 *
 * Extracts an item (either Copy or Inflate) from the zipfile to a FileDescriptior.
 * @param item: points to item to be extracted.
 * @param outFD: filedescriptor. May be 0, in which case file is not written, but only tested.
 */
PRInt32
nsZipArchive::ExtractToFileDesc(const nsZipItem* item, PRFileDesc* outFD)
{
  //-- sanity check arguments
  if ( item == 0 )
    return ZIP_ERR_PARAM;

  //-- extract the file using the appropriate method
  switch( item->GetCompression() )
  {
  case STORED: return CopyItemToFileDesc( item, outFD );
  case DEFLATED: return InflateToFileDesc( item, outFD );
  }
  return ZIP_ERR_UNSUPPORTED;
}

/*
 * nsZipArchive::CopyItemToFileDesc
 *
 * aItem points to the item to be copied.
 * fOut is a filedescriptor. fOut may be 0, in which case the item
 * is inflated but not written to file, for testing purposes.
 */
PRInt32 nsZipArchive::CopyItemToFileDesc(const nsZipItem* item, PRFileDesc* fOut)
{
  PR_ASSERT( item != 0 );

  //-- copy chunks until file is done
  PRInt32 offset = GetOffset( item );
  PRUint32 size = item->GetSize();

  //-- Calculate crc32
  PRUint 32 crc = crc32(0L, Z_NULL, 0);
  crc = crc32(crc, (const unsigned char*)(mMap + offset), size);
  if ( (fOut != 0) && (PR_Write( fOut, mMap + offset, size) < (READTYPE)size) )
      return ZIP_ERR_DISK;

  //-- verify crc32
  if ( (status == ZIP_OK) && (crc != item->GetCRC32()) )
      return ZIP_ERR_CORRUPT;

  return ZIP_OK;
}

/*
 * nsZipArchive::InflateToFileDesc
 *
 * aItem points to the item to be inflated.
 * fOut is a filedescriptor. fOut may be 0, in which case the item
 * is inflated but not written to file, for testing purposes.
 */
PRInt32 nsZipArchive::InflateToFileDesc( const nsZipItem* aItem, PRFileDesc* fOut)
{
  PRInt32     status = ZIP_OK;    
  z_stream    zs;
  int         zerr;
  Bytef       outbuf[ZIP_BUFLEN];

  PR_ASSERT( aItem != 0 );

  //-- set up the inflate
  memset( &zs, 0, sizeof(zs) );

#ifndef STANDALONE
  //-- ensure we have our zlib allocator for better performance
  if (!gZlibAllocator) {
    gZlibAllocator = new nsRecyclingAllocator(NBUCKETS, NS_DEFAULT_RECYCLE_TIMEOUT, "libjar");
  }
  zs.zalloc = zlibAlloc;
  zs.zfree = zlibFree;
  zs.opaque = gZlibAllocator;
#endif

  zerr = inflateInit2( &zs, -MAX_WBITS );
  if ( zerr != Z_OK )
  {
    status = ZIP_ERR_GENERAL;
    goto cleanup;
  }
  PRUint32 realSize = aItem->GetRealSize();
  PRUint32 size = aItem->GetSize();
  PRUint32 outsize = 0;
  PRInt32 offset = GetOffset( aItem );
  PRUint32 crc = crc32(0L, Z_NULL, 0);

  zs.next_in  = mMap + offset;
  zs.avail_in = size;

  //-- inflate loop
  while ( zerr == Z_OK )
  {
    // Write in blocks of ZIP_BUFLEN
    zs.next_out = outbuf;
    zs.avail_out = ZIP_BUFLEN;

    zerr = inflate( &zs, Z_SYNC_FLUSH );
    if (zerr != Z_OK && zerr != Z_STREAM_END)) {
      status = (zerr == Z_MEM_ERROR) ? ZIP_ERR_MEMORY : ZIP_ERR_CORRUPT;
      break;
    }

    PRUint32 chunk = ZIP_BUFLEN - zs.avail_out;
    if (chunk > 0 && (zerr == Z_OK || zerr == Z_STREAM_END)) {

      // Check output length 
      outsize += chunk;
      if ( outsize > realSize )
      {
        status = ZIP_ERR_CORRUPT;
        break;
      }

      //-- incrementally update crc32
      crc = crc32(crc, (const unsigned char*)outbuf, chunk);

      //-- write inflated buffer to disk and make space
      if ( (fOut != 0) && (PR_Write( fOut, outbuf, chunk ) < chunk) )
      {
        //-- Couldn't write all the data (disk full?)
        status = ZIP_ERR_DISK;
        break;
      }
    }    
#if defined STANDALONE && defined XP_WIN
    ProcessWindowsMessages();
#endif
  } // while

  //-- verify crc32
  if ( (status == ZIP_OK) && (crc != aItem->GetCRC32()) )
    status = ZIP_ERR_CORRUPT;

  //-- if found no errors make sure we've converted the whole thing
  PR_ASSERT( status != ZIP_OK || zs.total_in == aItem->GetSize() );
  PR_ASSERT( status != ZIP_OK || zs.total_out == aItem->GetRealSize() );

cleanup:
  //-- free zlib internal state
  inflateEnd( &zs );

  return status;
}


//---------------------------------------------
// nsZipArchive::InflateToBuffer
//---------------------------------------------
PRInt32 nsZipArchive::InflateToBuffer( const nsZipItem* aItem, char * outBuffer)
{
  PRInt32     status = ZIP_OK;    
  z_stream    zs;
  int         zerr;

  PR_ASSERT( aItem != 0 && fOut !=0 );

  //-- set up the inflate
  memset( &zs, 0, sizeof(zs) );

#ifndef STANDALONE
  //-- ensure we have our zlib allocator for better performance
  if (!gZlibAllocator) {
    gZlibAllocator = new nsRecyclingAllocator(NBUCKETS, NS_DEFAULT_RECYCLE_TIMEOUT, "libjar");
  }

  zs.zalloc = zlibAlloc;
  zs.zfree = zlibFree;
  zs.opaque = gZlibAllocator;
#endif

  zerr = inflateInit2( &zs, -MAX_WBITS );
  if ( zerr != Z_OK )
  {
    status = ZIP_ERR_GENERAL;
  }
  else
  {
    PRUint32 realSize = aItem->GetRealSize();
    zs.next_in  = mMap + GetOffset( aItem );
    zs.avail_in = aItem->GetSize();
    zs.next_out = outBuffer;
    zs.avail_out = realSize;
    
    zerr = inflate( &zs, Z_FINISH );
    if (zerr != Z_OK && zerr != Z_STREAM_END)) {
      status = (zerr == Z_MEM_ERROR) ? ZIP_ERR_MEMORY : ZIP_ERR_CORRUPT;
    }
  
    //-- if found no errors make sure we've converted the whole thing
    PR_ASSERT( status != ZIP_OK || zs.total_in == aItem->GetSize() );
    PR_ASSERT( status != ZIP_OK || zs.total_out == aItem->GetRealSize() );
  }
cleanup:
  //-- free zlib internal state
  inflateEnd( &zs );

  return status;
}


//---------------------------------------------
// nsZipArchive::TestItem
//---------------------------------------------
PRInt32 nsZipArchive::TestItem( const nsZipItem* aItem )
{
  // Test by extracting to a zero FileDescriptor (like /dev/null...)
  return ExtractToFileDesc( aItem, 0);
}

//------------------------------------------
// nsZipArchive constructor and destructor
//------------------------------------------

MOZ_DECL_CTOR_COUNTER(nsZipArchive)

nsZipArchive::nsZipArchive()
  : kMagic(ZIP_MAGIC), mMap(0), mItems(0), count(0)
{
  MOZ_COUNT_CTOR(nsZipArchive);

  // initialize the table to NULL
  memset(mFiles, 0, sizeof mFiles);
}

nsZipArchive::~nsZipArchive()
{
  (void)CloseArchive();

  MOZ_COUNT_DTOR(nsZipArchive);
}

//------------------------------------------
// nsZipItem constructor and destructor
//------------------------------------------

nsZipItem::nsZipItem() : next(0), central(0)
{
}

nsZipItem::~nsZipItem()
{
}

//------------------------------------------
// nsZipRead
//------------------------------------------

MOZ_DECL_CTOR_COUNTER(nsZipRead)

//------------------------------------------
// nsZipFind constructor and destructor
//------------------------------------------

MOZ_DECL_CTOR_COUNTER(nsZipFind)

nsZipFind::nsZipFind( nsZipArchive* aZip, char* aPattern, PRBool aRegExp )
: kMagic(ZIPFIND_MAGIC), 
  mArchive(aZip), 
  mPattern(aPattern), 
  mSlot(0), 
  mItem(0), 
  mRegExp(aRegExp) 
{
  MOZ_COUNT_CTOR(nsZipFind);
}

nsZipFind::~nsZipFind()
{
  if (mPattern != 0 )
    PL_strfree( mPattern );

  MOZ_COUNT_DTOR(nsZipFind);
}

//------------------------------------------
// nsZipFind::GetArchive
//------------------------------------------
nsZipArchive* nsZipFind::GetArchive()
{
    if (!mArchive)
        return NULL;

    return mArchive;
}


//------------------------------------------
// nsZipItem::GetName
//------------------------------------------
PRInt32  
nsZipItem::GetName( char * outbuf, PRUint16 bufsize )
{
  PRUint16 namelen = xtoint( central->filename_len );
  if ( bufsize < namelen ) 
    return ZIP_ERR_SMALLBUF;
  memcpy( outbuf, central + ZIPCENTRAL_SIZE, namelen);
  outbuf[namelen] = 0;
  return ZIP_OK;
}

//------------------------------------------
// nsZipItem::GetNameDup
//------------------------------------------
char *
nsZipItem::GetNameDup()
{
  PRUint16 namelen = xtoint( central->filename_len );
  char *namedup = (char *)malloc(namelen+1 * sizeof(char));
  if ( !namedup )
    return 0;
 
  memcpy(namedup, central + ZIPCENTRAL_SIZE, namelen);
  namedup[namelen] = 0;
  return namedup;
}

PRUint16 
nsZipItem::GetCompression() {
    return xtoint( central->method );
}
PRUint32
nsZipItem::GetRealSize() {
    return xtolong( central->orglen );
}
PRUint32
nsZipItem::GetSize() {
    return xtolong( central->size );
}
PRUint32
nsZipItem::GetCRC32() {
    return xtolong( central->crc32 );
}

#ifdef XP_UNIX
/*
 * GetMode
 *
 * Extracts bits 17-24 from a 32-bit unsigned long
 * representation of the external attributes field.
 * Subsequently it tacks on the implicit user-read
 * bit.
 */
PRUint16
nsZipItem::GetMode()
{
    unsigned char *ll = central->external_attributes;
    return ((PRUint16) ll[2]) | 0x00000100;
}

/*
 *  Return true if the attributes are for a symbolic link 
 */
PRBool
nsZipItem::IsSymLink() {
    PRUint32 ext = xtolong(central->external_attributes);
    return (((ext_attr>>16) & S_IFMT) == S_IFLNK) ? PR_TRUE : PR_FALSE;
}

//---------------------------------------------
// nsZipArchive::ResolveSymlink
//---------------------------------------------
PRInt32 
nsZipArchive::ResolveSymlink(const char *path) 
{
  PRInt32    status = ZIP_OK;
  char buf[PATH_MAX+1];
  PRFileDesc * fIn = PR_Open(path, PR_RDONLY, 0644);
  if (fIn) 
  {
    PRInt32 length = PATH_MAX;
    length = PR_Read(fIn,(void*)buf,length);
    PR_Close(fIn); 
    fIn = 0;
    if (   length <= 0
    || (buf[length] = 0, PR_Delete(path)) != 0
    || symlink(buf, path) != 0 )
    {
      status = ZIP_ERR_DISK;
    }
  } else {
    status = ZIP_ERR_DISK;
  }
  if (fIn) 
  {
    PR_Close(fIn);
  }
  return status;
}
#endif
 
/* 
 * Returns formatted date/time e.g. 06/20/1995 21:07
 * NSPR bug parsing dd/mm/yyyy hh:mm
 *        so we use mm/dd/yyyy hh:mm
 */
char *
nsZipItem::GetModTime()
{
    char *nsprstr = (char *) PR_Malloc(17 * sizeof(char));
    if (!nsprstr)
        return nsnull;
       
    PRUint16 aTime = xtoint( central->time );
    PRUint16 aDate = xtoint( central->date );

    sprintf(nsprstr, "%02d/%02d/%04d %02d:%02d",
        ((aDate >> 5) & 0x0F), (aDate & 0x1F), (aDate >> 9) + 1980);
        ((aTime >> 11) & 0x1F), ((aTime >> 5) & 0x3F));
    
    return nsprstr;
}
 
//------------------------------------------
// helper functions
//------------------------------------------

/*
 *  x t o i n t
 *
 *  Converts a two byte ugly endianed integer
 *  to our platform's integer.
 */
static PRUint16 xtoint (unsigned char *ii)
{
  return (PRUint16) ((ii [0]) | (ii [1] << 8));
}

/*
 *  x t o l o n g
 *
 *  Converts a four byte ugly endianed integer
 *  to our platform's integer.
 */
static PRUint32 xtolong (unsigned char *ll)
{
  PRUint32 ret;

  ret =  (
         (((PRUint32) ll [0]) <<  0) |
         (((PRUint32) ll [1]) <<  8) |
         (((PRUint32) ll [2]) << 16) |
         (((PRUint32) ll [3]) << 24)
         );

  return ret;
}

